#!/usr/bin/env python3
# encoding: utf-8
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2024 GEOMAR
# Copyright (C) 2022 Jakob Eckstein
# Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2023 Henrik tom Wörden <h.tomwoerden@indiscale.com>
# Copyright (C) 2024 Florian Spreckelsen <f.spreckelsen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import json
import logging
import os
import pandas as pd
import re
import subprocess
import sys

from datetime import date, datetime
from dateutil.relativedelta import relativedelta
from pathlib import Path
from tempfile import NamedTemporaryFile

import linkahead as db

from caosadvancedtools.datainconsistency import DataInconsistencyError
from caosadvancedtools.serverside import helper
from caosadvancedtools.table_importer import CSVImporter
from caoscrawler.logging import configure_server_side_logging

from bis_utils import (replace_entity_urls_by_ids,
                       SPECIAL_TREATMENT_SAMPLE, whitespace_cleanup_in_df)

# suppress warning of diff function
apilogger = logging.getLogger("linkahead.apiutils")
apilogger.setLevel(logging.ERROR)


def semicolon_separated_list(text):
    return [el.strip() for el in text.split(";") if el != ""]


def _embargo_converter(text: str):

    datepattern = r"^(?P<year>\d{4,4})-(?P<month>\d{2,2})-(?P<day>\d{2,2})"
    matches = re.match(datepattern, str(text))
    if matches:
        return date(int(matches.groupdict()["year"]), int(matches.groupdict()["month"]), int(matches.groupdict()["day"]))
    if f"{text}".lower() in ["true", "yes"]:
        # yes means embargo until today in one year
        return date.today() + relativedelta(years=1)
    if f"{text}".lower() in ["false", "no"]:
        return ""
    raise ValueError(
        f"The embargo should be either a date in YYYY-MM-DD format, or 'true'/'yes' or 'false'/'no', but is {text}.")


ERROR_PREFIX = 'Something went wrong: '
ERROR_SUFFIX = ' Please conatct <a href="mailto:biosamples@geomar.de">biosamples@geomar.de</a> if you encounter this issue.'


# Column datatypes
DATATYPE_DEFINITIONS = {
    "AphiaID": int,
    "BIS ID": str,
    "Campaign": str,
    "Date collected start": str,
    "Date collected stop": str,
    "Date sampled start": str,
    "Date sampled stop": str,
    "Fixation": str,
    "Gear configuration": str,
    "Gear": str,
    "Hol": int,
    "Latitude start": float,
    "Latitude stop": float,
    "Longitude start": float,
    "Longitude stop": float,
    "Main User": str,
    "Nagoya case number": str,
    "PI": str,
    "Parent BIS ID": str,
    "Platform": str,
    "Sample Context": str,
    "Sample container": str,
    "SampleType": str,
    "SampleTypeSpecific": str,
    "Sampling Person": str,
    "Sampling depth start": float,
    "Sampling depth stop": float,
    "Sampling method": str,
    "Station ID": str,
    "Station number": str,
    "Storage Container Label": str,
    "Storage ID": str,
    "StorageTemperature": str,
    "Subevent": str,
    "Time collected start": str,
    "Time collected stop": str,
    "Time sampled start": str,
    "Time sampled stop": str,
    "Timezone": str,
    "Water depth start": float,
    "Water depth stop": float,
}

# Obligatory columns: Must exist and must not be empty
# Must exist
OBLIGATORY_COLUMNS = [
    "BIS ID",
    "Collection",
    "Date collected start",
    "Fixation",
    "Gear",
    "Latitude start",
    "Longitude start",
    "Main User",
    "Nagoya case number",
    "PI",
    "Sample Context",
    "Sample container",
    "SampleType",
    "SampleTypeSpecific",
    "Sphere",
    "Storage ID",
    "StorageTemperature",
]

OBLIGATORY_COLUMNS_CHILD = [
    "BIS ID",
    "Date sampled start",
    "Fixation",
    "Main User",
    "Parent BIS ID",
    "Sample Context",
    "Sample container",
    "SampleType",
    "SampleTypeSpecific",
    "Sphere",
    "Storage ID",
    "StorageTemperature",
]

COLUMN_CONVERTER = {
    "Collection": semicolon_separated_list,
    "Ecotaxa URL": semicolon_separated_list,
    "NCBI Accession": semicolon_separated_list,
    "NCBI BioProject": semicolon_separated_list,
    "NCBI BioSample": semicolon_separated_list,
    "OSIS URL": semicolon_separated_list,
    "Embargo": _embargo_converter,
    "Publications": semicolon_separated_list,
    "Sphere": semicolon_separated_list,
}

logger = logging.getLogger("caosadvancedtools")


def get_parser():
    par = helper.get_argument_parser()
    return par


def _is_child_sample_table(filename):
    tmp_data = pd.read_csv(filename, sep=',')
    if 'Parent BIS ID' in tmp_data.columns:
        return not tmp_data["Parent BIS ID"].isnull().all()
    return False


def read_data_from_file(filename):
    if _is_child_sample_table(filename):
        oblig = OBLIGATORY_COLUMNS_CHILD
    else:
        oblig = OBLIGATORY_COLUMNS

    table_importer = CSVImporter(
        converters=COLUMN_CONVERTER,
        obligatory_columns=oblig,
        unique_keys=None,
        datatypes=DATATYPE_DEFINITIONS,
        existing_columns=oblig,
    )
    try:
        df = table_importer.read_file(filename, sep=",", comment="#")
    except TypeError as te:
        logger.error(
            f"There was a wrong datatype detected in your CSV: \n{te}\n"
            "Please verify that all entries have the correct type, e.g., no floating "
            "point numbers in integer columns or text in numeric columns."
        )
        raise DataInconsistencyError("There was a problem with the CSV upload.")

    # strip leading and trailing whitespaces
    return whitespace_cleanup_in_df(df)


def _get_converter_from_property_datatype(dt):
    if dt == db.TEXT:
        return str
    elif dt == db.REFERENCE:
        return int
    elif dt == db.DOUBLE:
        return float
    elif dt == db.INTEGER:
        return int
    elif dt == db.FILE:
        return int
    elif dt == db.DATETIME:
        raise NotImplementedError()
    elif dt == db.BOOLEAN:
        return bool
    else:
        raise ValueError(f"Property has unknown datatype {dt}")


def main():
    parser = get_parser()
    args = parser.parse_args()
    # Check whether executed locally or as an SSS depending on
    # auth_token argument.
    if hasattr(args, "auth_token") and args.auth_token:
        db.configure_connection(auth_token=args.auth_token)
        userlog_public, htmluserlog_public, debuglog_public = configure_server_side_logging()
    else:
        rootlogger = logging.getLogger()
        rootlogger.setLevel(logging.INFO)
        logger.setLevel(logging.DEBUG)
        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setLevel(logging.DEBUG)
        rootlogger.addHandler(handler)
        userlog_public = "/tmp/upload_sample_userlog.log"
        htmluserlog_public = "/tmp/upload_sample_userlog.html"
        debuglog_public = "/tmp/upload_sample_debuglog.html"

    if hasattr(args, "filename") and args.filename:
        if hasattr(args, "auth_token") and args.auth_token:
            upload_dir = os.path.dirname((args.filename))
            # Read the input from the form (form.json)
            with open(args.filename) as form_json:
                form_data = json.load(form_json)
            # Read content of th uplaoded file
            path = os.path.join(upload_dir, form_data["sample_template_file"])
        else:
            path = args.filename

        # Extend the converter lists by those properties that are unknown to this script but are
        # Properties of Sample
        sample = db.execute_query("FIND RECORDTYPE Sample", unique=True)
        for eprop in sample.properties:
            property_name = eprop.name
            if property_name in SPECIAL_TREATMENT_SAMPLE:
                continue
            if db.apiutils.is_reference(eprop):
                rt = db.get_entity_by_id(eprop.id)
                if len(rt.properties) == 1:
                    converter = _get_converter_from_property_datatype(rt.properties[0].datatype)
                elif len(rt.properties) == 1:
                    converter = str
                else:
                    converter = None
            else:
                converter = _get_converter_from_property_datatype(eprop.datatype)
            if converter is None:
                continue
            DATATYPE_DEFINITIONS[property_name] = converter
            if sample.get_importance(property_name) == db.OBLIGATORY:
                # This is only needed if the sample is not a child sample
                OBLIGATORY_COLUMNS.append(property_name)
        try:
            data = read_data_from_file(path)
            data = replace_entity_urls_by_ids(data, ["BIS ID", "Storage ID", "Parent BIS ID"])
            pickle_out = NamedTemporaryFile(delete=False, suffix=".pkl")
            data.to_pickle(pickle_out.name)
        except DataInconsistencyError as err:
            # DataInconsistencyError is logged elsewhere
            logger.error(f"Sample upload failed.\n{err}")
            return 1

        parent_path = Path(__file__).parent
        db.Info()  # call db.Info to refresh the auth_token, just to be sure.
        conn = db.get_connection()
        auth_token_callee = conn._authenticator.auth_token

        cmds = [
            str(parent_path / "crawl_sample_data_async.py"),
            "--auth-token",
            auth_token_callee,
            args.filename,
            pickle_out.name,
            Path(path).name
        ]

        myenv = os.environ.copy()
        myenv["HOME"] = str(parent_path.parent / "home")

        # For a few samples, we can run the upload directly and give
        # user output. More samples will be run asynchronously.
        if data.shape[0] < 20:
            p = subprocess.Popen(cmds, start_new_session=False, env=myenv,
                                 cwd=parent_path, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()
            if p.returncode == 0:
                # We know that the stdout will be formatted by the
                # serverside logging helper, so we don't wrap it but
                # just print it.
                print(stdout.decode())
            else:
                logger.error(stderr.decode())
                print(stdout.decode())
        else:
            logger.info(
                "Starting sample upload in the background. This may take a while. "
                "You will be notified by email when it has finished."
            )
            p = subprocess.Popen(cmds, start_new_session=True, env=myenv,
                                 cwd=parent_path)

    else:
        msg = "{}upload_sample_template.py was called without the JSON file in args.{}".format(
            ERROR_PREFIX, ERROR_SUFFIX)
        logger.error(msg)


if __name__ == "__main__":
    main()
